import os
from pathlib import Path
from random import choices
from string import ascii_uppercase, digits

import httpx
import pytest
from click.testing import CliRunner

import hukudo.cli
from hukudo.grafana import Grafana, demo
from hukudo.grafana.errors import DashboardWriteError
from hukudo.grafana.models import JSONThing


@pytest.fixture(scope='session')
def random_string(k=6):
    return str.join('', choices(ascii_uppercase + digits, k=k))


@pytest.fixture(scope='session')
def api_key(random_string):
    return demo.api_key(f'test-{random_string}')


@pytest.fixture(scope='session')
def demo_env(api_key):
    return {
        'GRAFANA_URL': demo.URL,
        'GRAFANA_API_KEY': api_key,
        'GRAFANA_CLIENT_ROOT_CA': str(Path('~/ingress/root.crt').expanduser()),
    }


@pytest.fixture()
def demo_envfile(demo_env, tmp_path):
    config_file = tmp_path / 'config'
    with config_file.open('w') as f:
        for k, v in demo_env.items():
            f.write(f'{k}={v}\n')
    return config_file


@pytest.fixture(scope='session')
def grafana(demo_env, example_dashboard_json):
    return Grafana.from_api_key(
        demo_env['GRAFANA_URL'],
        demo_env['GRAFANA_API_KEY'],
        verify=demo_env['GRAFANA_CLIENT_ROOT_CA'],
    )


@pytest.fixture(autouse=True)
def auto_mock_patch_os_environ(mocker):
    """
    CLI calls may change os.environ, e.g. dotenv.load_dotenv().
    Mock os.environ using the current environment and restore afterwards.
    https://docs.python.org/3/library/unittest.mock.html#patch-dict
    """
    mocker.patch.dict('os.environ', os.environ)


@pytest.fixture(scope='session')
def example_dashboard_json():
    return {
        "annotations": {
            "list": [
                {
                    "builtIn": 1,
                    "datasource": {"type": "grafana", "uid": "-- Grafana --"},
                    "enable": True,
                    "hide": True,
                    "iconColor": "rgba(0, 211, 255, 1)",
                    "name": "Annotations & Alerts",
                    "target": {"limit": 100, "matchAny": False, "tags": [], "type": "dashboard"},
                    "type": "dashboard",
                }
            ]
        },
        "editable": True,
        "fiscalYearStartMonth": 0,
        "graphTooltip": 0,
        "links": [],
        "liveNow": False,
        "panels": [
            {
                "datasource": {"type": "datasource", "uid": "grafana"},
                "gridPos": {"h": 9, "w": 12, "x": 0, "y": 0},
                "id": 2,
                "options": {"content": "# Hello World!", "mode": "markdown"},
                "pluginVersion": "8.5.0",
                "title": "Test Panel",
                "type": "text",
            }
        ],
        "schemaVersion": 36,
        "style": "dark",
        "tags": [],
        "templating": {"list": []},
        "time": {"from": "now-6h", "to": "now"},
        "timepicker": {},
        "timezone": "",
        "title": "test",
        "uid": "test-uid",
        "version": 0,
        "weekStart": "",
    }


def test_json_thing():
    x = JSONThing({'foo': 'bar', 'baz': 123, 'qux': 'quux'})
    assert repr(x) == "{'foo': 'bar', 'baz': 123, 'qu"


def test_client_cert():
    from dotenv import dotenv_values

    env = dotenv_values()
    grafana = Grafana(
        httpx.Client(
            base_url=env['GRAFANA_URL'],
            headers=Grafana.bearer_header(env['GRAFANA_API_KEY']),
            verify=env['GRAFANA_CLIENT_ROOT_CA'],
            cert=(
                env['GRAFANA_CLIENT_CRT'],
                env['GRAFANA_CLIENT_KEY'],
            ),
        )
    )
    assert grafana.health()
    dashboards = grafana.dashboards()
    assert len(dashboards) > 10


def test_api_key(api_key):
    assert len(api_key) >= 80


def test_api_key_role_invalid():
    client = httpx.Client()
    with pytest.raises(ValueError):
        Grafana(client).provision_api_key('_', 'ThisIsInvalid')


def test_post_dashboard(grafana, example_dashboard_json):
    grafana.post_dashboard(example_dashboard_json)


def test_post_dashboard_error(grafana):
    with pytest.raises(DashboardWriteError):
        grafana.post_dashboard({'uid': 'foo'})


def run(env, args):
    runner = CliRunner()
    # noinspection PyTypeChecker
    result = runner.invoke(hukudo.cli.root, args, env=env, catch_exceptions=False)
    return result


def test_cli_health(demo_env):
    run(demo_env, ['grafana', 'health'])


def test_cli_dashboard_list(demo_env):
    result = run(demo_env, ['grafana', 'dashboard', 'list'])
    assert 'https://grafana.dev.0-main.de/d/test-uid' in result.output


def test_cli_dashboard_export_import(demo_env, tmp_path):
    run(demo_env, ['grafana', 'dashboard', 'export', str(tmp_path)])
    files = list(tmp_path.glob('*.json'))
    assert len(files) == 1
    # formatted json is indented on line 2
    contents = files[0].read_text()
    line_2 = contents.split('\n')[1]
    assert line_2.startswith('  ')
    for file in files:
        run(demo_env, ['grafana', 'dashboard', 'import', str(file)])


def test_cli_config(demo_envfile):
    run(env={}, args=['grafana', '-c', demo_envfile, 'dashboard', 'list'])


def test_cli_errors_empty_env(mocker, tmp_path):
    mocker.patch.dict('os.environ', {})  # empty os.environ
    os.chdir(tmp_path)  # make sure there is no .env file present
    runner = CliRunner(mix_stderr=False)
    # noinspection PyTypeChecker
    result = runner.invoke(hukudo.cli.root, ['grafana', 'health'], catch_exceptions=False)
    assert "missing environment variable 'GRAFANA_URL'" in result.stderr



import json

import click
import structlog

from hukudo.gitlab.api import Gitlab
from hukudo.gitlab.jobs import get_duration, JobDurationParseError

logger = structlog.get_logger()


@click.group()
@click.option('--name')
@click.pass_context
def gitlab(ctx, name):
    ctx.obj = Gitlab.from_ini(name)
    log = logger.bind(instance=ctx.obj)
    log.debug('instantiated')


@gitlab.command()
@click.argument('project')
@click.pass_context
def jobs(ctx, project):
    gitlab: Gitlab = ctx.obj
    for job in gitlab.jobs_of_project(project):
        try:
            job.attributes['duration'] = get_duration(job.attributes)
            print(json.dumps(job.attributes))
        except JobDurationParseError:
            pass



import time

import pytest

from hukudo.filesystem import keepn


@pytest.fixture
def dump_dir(tmp_path):
    for i in range(1, 45 + 1):
        filename = f'{i:02}.dump'
        tmp_path.joinpath(filename).touch()
        time.sleep(0.01)  # to get different modified times (mtime)
    return tmp_path


def test_e2e(dump_dir):
    assert len(list(dump_dir.glob('*'))) == 45

    keepn(dump_dir, n_keep=40)
    files = list(dump_dir.glob('*'))
    assert len(files) == 40
    # first few get removed
    assert '01.dump' not in str(files)
    assert '05.dump' not in str(files)
    # rest stays
    assert '06.dump' in str(files)
    assert '45.dump' in str(files)



from datetime import datetime, timedelta


class JobDurationParseError(ValueError):
    pass


def get_duration(attributes) -> timedelta:
    try:
        started_at = datetime.fromisoformat(attributes['started_at'])
        finished_at = datetime.fromisoformat(attributes['finished_at'])
        return finished_at - started_at
    except (KeyError, ValueError, TypeError):
        raise JobDurationParseError('could not parse')



from datetime import timedelta

import pytest

from hukudo.gitlab.jobs import get_duration, JobDurationParseError


def test_job_with_duration_happy():
    assert get_duration(
        {
            'started_at': '2022-07-19T16:13:17.374+02:00',
            'finished_at': '2022-07-19T16:14:19.374+02:00',
        }
    ) == timedelta(seconds=62)


def test_job_with_duration_error():
    with pytest.raises(JobDurationParseError):
        get_duration(
            {
                'started_at': 'not-an-iso-datetime-string',
                'finished_at': 'not-an-iso-datetime-string',
            }
        )



from __future__ import annotations

import typing
from typing import Union

from openpyxl import utils as openpyxl_utils
from openpyxl.utils.cell import coordinate_from_string


class EDI(str):
    """
    Excel Document Iterator
    """

    def __new__(cls, coords):
        return super().__new__(cls, coords)

    def __repr__(self):
        return f"'{self.col}{self.row}'"

    @classmethod
    def from_col_row(cls, col: str, row: int):
        return cls(f'{col}{row}')

    @property
    def tuple(self) -> tuple[str, int]:
        """
        >>> EDI('B4').tuple
        ('B', 4)
        """
        return coordinate_from_string(self)

    @property
    def col(self):
        return self.tuple[0]

    @property
    def row(self):
        return self.tuple[1]

    @property
    def col_num(self):
        """
        >>> EDI('B4').col_num
        2
        """
        return openpyxl_utils.column_index_from_string(self.col)

    def at_row(self, i) -> EDI:
        return self.__class__(f'{self.col}{i}')

    def at_col(self, col):
        """
        >>> EDI('B23').at_col('A')
        'A23'
        """
        return self.from_col_row(col, self.row)

    def row_offset(self, i) -> EDI:
        """
        >>> EDI('B4').row_offset(2)
        'B6'
        """
        return self.at_row(self.row + i)

    def col_offset(self, i: int) -> EDI:
        """
        >>> EDI('B4').col_offset(2)
        'D4'
        """
        return self.at_col(openpyxl_utils.get_column_letter(self.col_num + i))

    def with_row_indices(self, values: typing.Iterable[str]) -> dict:
        """
        >>> EDI('B4').with_row_indices(['foo', 'bar', 'baz'])
        {'B4': 'foo', 'B5': 'bar', 'B6': 'baz'}
        """
        result = {}
        for i, value in enumerate(values):
            result[self.row_offset(i)] = value
        return result

    def with_col_indices(self, values: typing.Iterable[str]) -> dict:
        """
        >>> EDI('B4').with_col_indices(['foo', 'bar', 'baz'])
        {'B4': 'foo', 'C4': 'bar', 'D4': 'baz'}
        """
        result = {}
        for i, value in enumerate(values):
            result[self.col_offset(i)] = value
        return result

    def get_range_col_offset(self, i: int):
        """
        >>> EDI('B4').get_range_col_offset(2)
        'B4:D4'
        """
        return f'{self}:{self.col_offset(i)}'


EDISTR = Union[EDI, str]



import logging

from hukudo.xls import TableError

log = logging.getLogger(__name__)


class Table:
    def __init__(self, header: list, data: list[list]):
        self.header = header
        self.data = data

    def __bool__(self):
        """
        Tables are False if they have no data. Useful for checks like "if table: ..."
        """
        return bool(self.data)

    def dicts(self):
        result = []
        for d in self.data:
            result.append({k: v for (k, v) in zip(self.header, d)})
        return result

    def single_dict(self):
        dicts = self.dicts()
        if len(dicts) != 1:
            raise TableError('More than one result found. Please fix your query to return exactly one row.')
        return dicts[0]

    def change_columns(self, mapping: dict = None, capitalize=True):
        """
        Siehe test case.
        """
        if mapping is None:
            mapping = {}
        # header id to new value
        update_names = {self.header.index(k): v for k, v in mapping.items() if v is not None}
        # header ids to drop
        drop_indices = [self.header.index(k) for k, v in mapping.items() if v is None]
        # header id to new value
        capitalize_names = {}
        if capitalize:
            for i, name in enumerate(self.header):
                new_name = name
                if name == 'id':
                    new_name = 'ID'
                elif name not in mapping:
                    new_name = name.capitalize()
                if name != new_name:
                    capitalize_names[i] = new_name
        # last one wins. mapping has prio.
        update_rules = {}
        update_rules.update(capitalize_names)
        update_rules.update(update_names)
        for k, v in update_rules.items():
            self.header[k] = v
        self.header = [x for i, x in enumerate(self.header) if i not in drop_indices]
        for i, row in enumerate(self.data):
            self.data[i] = [x for i, x in enumerate(row) if i not in drop_indices]



import logging
import os
from pathlib import Path

import click
import structlog
from click import ClickException

import hukudo
from hukudo.exc import HukudoException
from hukudo.log import LOG_LEVELS, configure_structlog_dev
from .gitlab.cli import gitlab
from .grafana.cli import grafana
from .pypi.cli import pypi


@click.group()
@click.option(
    '-l',
    '--log-level',
    type=click.Choice(LOG_LEVELS, case_sensitive=False),
    default=lambda: os.environ.get('LOGLEVEL', 'WARNING'),
)
def root(log_level):
    """
    For completion, add this to ~/.bashrc:

        eval "$(_HUKUDO_COMPLETE=bash_source hukudo)"

    See also https://click.palletsprojects.com/en/8.1.x/shell-completion/
    """
    configure_structlog_dev(log_level)
    logging.basicConfig(format='%(msg)s')
    structlog.getLogger().debug('logging configured', level=log_level)


# noinspection PyTypeChecker
root.add_command(grafana)
# noinspection PyTypeChecker
root.add_command(gitlab)
# noinspection PyTypeChecker
root.add_command(pypi)


@root.command()
def version():
    print(hukudo.__version__)


@root.group()
def chromedriver():
    pass


@chromedriver.command()
@click.option('-s', '--skip-symlink', is_flag=True)
@click.option('-f', '--force', is_flag=True)
@click.argument('target_dir', type=click.Path(path_type=Path), required=False)
def download(skip_symlink, force, target_dir):
    """
    Downloads the latest chromedriver matching your Chrome browser version
    and creates a symlink to it.

    Example:

        hukudo chromedriver download /tmp/

    Results in `/tmp/chromedriver` pointing to `/tmp/chromedriver-101.0.4951.41`.
    """
    import hukudo.chromedriver

    try:
        hukudo.chromedriver.download_and_symlink(target_dir, force=force, skip_symlink=skip_symlink)
    except HukudoException as e:
        raise ClickException(str(e))


@root.command(name='keepn')
@click.argument('num', type=int)
@click.argument('directory', type=click.Path(path_type=Path))
def keepn_(num: int, directory: Path):
    """Keep NUM files in DIRECTORY and delete the rest."""
    import hukudo.filesystem

    hukudo.filesystem.keepn(directory=directory, n_keep=num)


def main():
    root()


if __name__ == '__main__':
    main()



from hukudo.grafana import Grafana

URL = 'https://grafana.dev.0-main.de'
USERNAME = 'admin'
PASSWORD = 'test'


def api_key(name):
    grafana = Grafana.from_basic_auth(URL, USERNAME, PASSWORD)
    return grafana.provision_api_key(name=name, role=Grafana.ADMIN)



from copy import copy

from openpyxl.styles import Font, PatternFill, Border, Side, DEFAULT_FONT, NamedStyle, Alignment
from openpyxl.styles.borders import DEFAULT_BORDER
from openpyxl.styles.fills import FILL_SOLID

# OVERRIDE default font
DEFAULT_FONT.size = 10
DEFAULT_FONT.name = 'Calibri'

align_left = Alignment(horizontal='left')

fat_top_border = copy(DEFAULT_BORDER)
fat_top_border.top = Side(border_style="thick", color="575757")
thin_top_border = copy(DEFAULT_BORDER)
thin_top_border.top = Side(border_style="thin", color="575757")

h1 = NamedStyle(name='h1')
h1.fill = PatternFill(fill_type=FILL_SOLID, start_color='d9d9d9', end_color='d9d9d9')
h1.font = copy(DEFAULT_FONT)
h1.font.color = '575757'
h1.font.size = 14
h1.font.bold = True
h1.border = copy(fat_top_border)

h1sub = NamedStyle(name='h1-sub')
h1sub.fill = copy(h1.fill)
h1sub.font = copy(h1.font)
h1sub.font.size = 11
h1sub.font.bold = False

h2 = NamedStyle(name='h2')
h2.fill = copy(h1.fill)
h2.font = copy(h1.font)
h2.font.size = 12
h2.border = thin_top_border

h2sub = NamedStyle(name='h2-sub')
h2sub.fill = copy(h2.fill)
h2sub.font = copy(h2.font)
h2sub.font.bold = True

h3 = NamedStyle(name='h3')
h3.fill = copy(h2.fill)
h3.font = copy(h2.font)
h3.font.size = 10
h3.border = thin_top_border

h3sub = NamedStyle(name='h3-sub')
h3sub.fill = copy(h3.fill)
h3sub.font = copy(h3.font)
h3sub.font.bold = False


TABLE_HEADER_FILL = PatternFill(fill_type=FILL_SOLID, start_color='575757', end_color='575757')
TABLE_HEADER_FONT = Font('Calibri', size=10, bold=True, color='FFFFFF')
TABLE_FILL_ALTERNATE = PatternFill(fill_type=FILL_SOLID, start_color='EEEEEE', end_color='EEEEEE')
HEADER_FONT = Font('Calibri', size=10, bold=True)
thin = Side(border_style="thin", color="b0b0b0")
BORDER = Border(top=thin, left=thin, right=thin, bottom=thin)
black = Side(border_style="thin", color="000000")
BORDER_BLACK = Border(top=black, left=black, right=black, bottom=black)

FMT_GENERAL = 'General'
FMT_EUR = '#,##0.00 €'
FMT_HOUR = '#,##0.00" h"'
FMT_KG = '#,##0.00" kg"'
FMTS = {
    'Arbeit': FMT_EUR,
    'Aufwand': FMT_HOUR,
    'Dauer': FMT_HOUR,
    'Einzelpreis': FMT_EUR,
    'Ges.gew.': FMT_KG,
    'Ges.preis': FMT_EUR,
    'Gesamt': FMT_EUR,
    'Gesamtgewicht': FMT_KG,
    'Gesamtpreis': FMT_EUR,
    'Gewicht': FMT_KG,
    'Material': FMT_EUR,
    'Materialpreis': FMT_EUR,
    'fixer Aufwand': FMT_HOUR,
    'var. Aufwand': FMT_HOUR,
}



import json
import os
from pathlib import Path

import click
import dotenv
import structlog

from hukudo.grafana import Grafana
from hukudo.grafana.errors import DashboardWriteError

logger = structlog.get_logger()


@click.group()
@click.option('-c', '--config', type=click.Path())
@click.pass_context
def grafana(ctx, config):
    if config:
        logger.debug('dotenv override', config=config)
        os.environ.update(dotenv.dotenv_values(config))
    # get stuff from env
    try:
        url = os.environ['GRAFANA_URL']
        api_key = os.environ['GRAFANA_API_KEY']
    except KeyError as e:
        raise click.ClickException(f'missing environment variable {e}')
    root_ca = os.environ.get('GRAFANA_CLIENT_ROOT_CA')
    logger.debug('CA', path=root_ca)

    client_cert = None
    try:
        logger.debug('read config', what='client_cert')
        crt = os.environ['GRAFANA_CLIENT_CRT']
        key = os.environ['GRAFANA_CLIENT_KEY']
        logger.debug('client cert', crt=crt, key=key)
        client_cert = (crt, key)
    except KeyError:
        pass

    basic_auth = None
    try:
        logger.debug('read config', what='basic_auth')
        username = os.environ['GRAFANA_BASIC_AUTH_USERNAME']
        password = os.environ['GRAFANA_BASIC_AUTH_PASSWORD']
        logger.debug('basic_auth', username=username)
        basic_auth = (username, password)
    except KeyError:
        pass

    ctx.obj = Grafana.from_api_key(
        url=url,
        api_key=api_key,
        verify=root_ca,
        cert=client_cert,
        auth=basic_auth,
    )
    log = logger.bind(instance=ctx.obj)
    log.debug('instantiated')


@grafana.command()
@click.pass_context
def health(ctx):
    grafana: Grafana = ctx.obj
    if grafana.health():
        logger.info('health OK', instance=grafana)


@grafana.group()
def dashboard():
    pass


@dashboard.command(name='list')
@click.pass_context
def dashboard_list(ctx):
    grafana: Grafana = ctx.obj

    for board in grafana.dashboards():
        print(f'{board.url}\t({board.title})')


@dashboard.command()
@click.argument('target_dir', type=click.Path())
@click.pass_context
def export(ctx, target_dir):
    """
    Exports all dashboards as JSON files to the given directory.
    """
    log = logger.bind(instance=ctx.obj)
    log.info('export')

    grafana: Grafana = ctx.obj
    target = Path(target_dir)

    for board in grafana.dashboards():
        filename = target / f'{board.id}.json'
        board.export(filename)


@dashboard.command(name='import')
@click.argument('paths', type=click.Path(), nargs=-1)
@click.pass_context
def import_(ctx, paths):
    """
    Import dashboards from JSON files.
    """
    grafana: Grafana = ctx.obj

    exit_code = 0
    for path in paths:
        path = Path(path)
        with path.open() as f:
            data = json.load(f)
        try:
            grafana.post_dashboard(data)
        except DashboardWriteError:
            exit_code = 1
    return exit_code



from click.testing import CliRunner

import hukudo.cli


def test_logs_to_stderr():
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(hukudo.cli.root, ['version'], env={'LOGLEVEL': 'debug'})
    assert hukudo.__version__ in result.stdout
    assert 'logging configured' in result.stderr



def test_change_columns(ws, table):
    # drop second column and rename last column
    table.change_columns(
        {
            'col2': None,
            'col3': 'Column Three',
        }
    )
    # note that change_columns automatically capitalizes column names
    assert table.header == ['Col1', 'Column Three', 'ID']
    assert table.data == [
        ['first', 'data', 'id data'],
        ['second', 'data', 'id data'],
    ]



import string
from pathlib import Path

import openpyxl.utils
from openpyxl.drawing.image import Image
from openpyxl.worksheet.worksheet import Worksheet

from . import fmt
from .edi import EDISTR, EDI
from .table import Table


class Sheet:
    def __init__(self, ws: Worksheet, fmts: dict = None):
        self.ws = ws
        self.fmts = fmts
        if self.fmts is None:
            self.fmts = fmt.FMTS

    def add_table(self, start: EDISTR, table: Table, fmts: dict = None) -> EDI:
        if fmts is None:
            fmts = {}
        start = EDI(start)
        my_formats = self.fmts | fmts
        formats = [my_formats.get(c) for c in table.header]
        for edi, v in start.with_col_indices(table.header).items():
            self.ws[edi] = v
            self.ws[edi].font = fmt.TABLE_HEADER_FONT
            self.ws[edi].fill = fmt.TABLE_HEADER_FILL
            self.ws[edi].border = fmt.BORDER
        for row_i, row in enumerate(table.data):
            stripe_row = row_i % 2 == 0
            data_start = start.row_offset(row_i + 1)  # +1 to jump over header
            for col_i, (edi, v) in enumerate(data_start.with_col_indices(row).items()):
                self.ws[edi] = v
                self.ws[edi].border = fmt.BORDER
                if stripe_row:
                    self.ws[edi].fill = fmt.TABLE_FILL_ALTERNATE
                x = formats[col_i]
                if x is not None:
                    self.ws[edi].number_format = x
        return start.row_offset(len(table.data) + 1)

    def add_table_top_down(self, start: EDISTR, table: Table, merge_cols: int = 2) -> EDI:
        """
        :merge_cols: Number of columns to merge for values.
        :return: Row index after inserted data, e.g. B11.
        """
        start = EDI(start)
        d = table.single_dict()
        # iterate rows
        for i, (key, value) in enumerate(d.items()):
            header_edi = start.row_offset(i)
            value_edi = start.row_offset(i).col_offset(1)
            self.ws[header_edi] = key
            self.ws[header_edi].font = fmt.TABLE_HEADER_FONT
            self.ws[header_edi].fill = fmt.TABLE_HEADER_FILL
            self.ws[header_edi].border = fmt.BORDER
            self.ws[value_edi] = value
            self.ws[value_edi].border = fmt.BORDER
            self.ws[value_edi].number_format = self.fmts.get(key, fmt.FMT_GENERAL)
            self.ws[value_edi].alignment = fmt.align_left
            # merge cells
            self.ws.merge_cells(value_edi.get_range_col_offset(merge_cols))
        return start.row_offset(len(d))

    def add_table_with_header(
        self, start: EDISTR, header: str, table: Table, style: str, end_col='J', fmts: dict = None
    ):
        start = EDI(start)
        self.ws[start] = header
        self.style_row(start, style, end_col)
        return self.add_table(start.row_offset(1), table, fmts=fmts)

    # noinspection DuplicatedCode
    def style_row(self, start: EDISTR, style: str, end_col: str = 'J'):
        start = EDI(start)
        a, b = string.ascii_uppercase.index(start.col), string.ascii_uppercase.index(end_col) + 1
        cols = string.ascii_uppercase[a:b]
        for c in cols:
            self.ws[start.at_col(c)].style = style
        return start.row_offset(1)

    # noinspection DuplicatedCode
    def fill_row(self, start: EDISTR, end_col: str, fill):
        start = EDI(start)
        a, b = string.ascii_uppercase.index(start.col), string.ascii_uppercase.index(end_col)
        cols = string.ascii_uppercase[a:b]
        for c in cols:
            self.ws[start.at_col(c)].fill = fill

    def auto_set_column_width(self, row_start=1, min_width=10):
        """
        https://stackoverflow.com/a/58667753/241240
        """
        for i, column_cells in enumerate(self.ws.columns, start=1):
            # note that black formats this weirdly. Just live with it. :P
            width = (
                length
                if (
                    length := max(
                        len(str(cell_value) if (cell_value := cell.value) is not None else "")
                        for cell in column_cells[(row_start - 1) :]  # noqa
                    )
                )
                >= min_width
                else min_width
            )
            # + 1 wegen units (kg)
            self.ws.column_dimensions[openpyxl.utils.get_column_letter(i)].width = width + 4

    def add_logo(self, path: Path):
        img = Image(path)
        img.anchor = 'B1'
        img.height = 79
        img.width = 162
        self.ws.add_image(img)

    def width_in_inch(self, column, inches):
        x = inches * 15.8
        self.ws.column_dimensions[column].width = x

    def set_widths_in_inch(self, col2inch: dict[str, float]):
        for col, inch in col2inch.items():
            self.width_in_inch(col, inch)



import os
import subprocess

import pytest
from click.testing import CliRunner

import hukudo.cli
from hukudo.exc import HukudoException


@pytest.fixture(autouse=True, scope='session')
def auto_requests_cache(session_mocker, tmp_path_factory):
    env = os.environ.copy()
    env['REQUESTS_CACHE_DIR'] = str(tmp_path_factory.mktemp('cache'))
    session_mocker.patch.dict('os.environ', env)


def test_download(tmp_path):
    from hukudo.chromedriver import download_and_symlink, is_matching_my_major

    # download and test-run the executable
    version, driver_path = download_and_symlink(tmp_path)
    output = subprocess.check_output([str(driver_path), '--version'], encoding='utf-8')
    assert 'ChromeDriver' in output
    assert str(version) in output
    # check the symlink
    link = tmp_path / 'chromedriver'
    assert link.absolute().is_file()
    assert is_matching_my_major(driver_path)


def test_lookup():
    from hukudo.chromedriver import ChromedriverRepo

    repo = ChromedriverRepo('linux64')
    versions = repo.versions()
    assert len(versions) > 0
    lv = repo.latest_version(100)
    assert lv.version.startswith('100.')


def test_my_chrome():
    from hukudo.chromedriver import my_chrome_version

    v = my_chrome_version()
    assert v.version
    assert v.major > 90


def run(args):
    runner = CliRunner(mix_stderr=False)
    # noinspection PyTypeChecker
    result = runner.invoke(hukudo.cli.root, args)
    return result


def test_cli(pytester):
    pytester.chdir()

    result = run(['chromedriver', 'download'])
    assert result.exit_code == 0
    path = next(pytester.path.glob('chromedriver-*'))
    assert path.is_file()
    x = path.stat()
    assert x.st_size > 0

    cannot_overwrite = run(['chromedriver', 'download'])
    assert 'Error: file exists' in cannot_overwrite.stderr
    assert cannot_overwrite.exit_code == 1

    can_force_overwrite = run(['chromedriver', 'download', '--force'])
    assert can_force_overwrite.stderr == ''

    skip_symlink = run(['chromedriver', 'download', '--force', '--skip-symlink'])
    assert skip_symlink.stderr == ''

    # must download to a directory
    file_path = pytester.path / 'foo'
    file_path.write_text('hello')
    # noinspection PyTypeChecker
    invalid_directory = run(['chromedriver', 'download', str(file_path)])
    assert 'Error: not a directory' in invalid_directory.stderr
    assert invalid_directory.exit_code == 1


def test_cli_error(mocker):
    # noinspection PyUnusedLocal
    def raiser(*args, **kwargs):
        raise HukudoException('test')

    mocker.patch('hukudo.chromedriver.download_and_symlink', raiser)
    runner = CliRunner(mix_stderr=False)
    # noinspection PyTypeChecker
    result = runner.invoke(hukudo.cli.root, ['chromedriver', 'download'])
    assert result.stderr == 'Error: test\n'



import pytest

from hukudo.xls import TableError, fmt
from hukudo.xls.sheet import Sheet


def test_add_table(sh, table):
    sh.add_table('B2', table)


def test_add_table_top_down(sh, table_single_row):
    sh.add_table_top_down('A1', table_single_row)


def test_add_table_top_down_error(sh, table):
    with pytest.raises(TableError):
        sh.add_table_top_down('A1', table)


def test_add_table_with_header(wb, table):
    # add the style "h2"
    wb.add_named_style(fmt.h2)
    sh = Sheet(wb.active)
    sh.add_table_with_header('A1', 'my table header', table, fmt.h2.name)


def test_fill_row(sh):
    sh.fill_row('A1', 'Z', fmt.h1.fill)


def test_auto_set_column_width(sh):
    sh.auto_set_column_width()
